__author__ = 'deadblue'

import json
import urllib

from py115._internal.crypto import m115
from py115._internal.protocol import api


class M115ApiSpec(api.ApiSpec):

    def __init__(self, url: str, use_ec: bool = False) -> None:
        super().__init__(url, use_ec)
        self._m_key = m115.generate_key()

    @property
    def payload(self) -> bytes:
        data = m115.encode(self._m_key, json.dumps(self._form))
        return urllib.parse.urlencode({
            'data': data
        }).encode()

    def parse_result(self, result: dict):
        # M115 decode
        data = m115.decode(
            self._m_key,  super().parse_result(result)
        )
        return json.loads(data)



__author__ = 'deadblue'

from py115._internal.protocol.api import ApiSpec

class GetVersionApi(ApiSpec):
    
    def __init__(self) -> None:
        super().__init__('https://appversion.115.com/1/web/1.0/api/chrome')

    def parse_result(self, result: dict) -> str:
        return result['data']['linux_115']['version_code']


class GetTypeApi(ApiSpec):

    def __init__(self) -> None:
        super().__init__('https://webapi.115.com/files/index_info')
    
    def parse_result(self, result: dict):
        result = super().parse_result(result)
        login_devices = result['login_devices_info']['list']
        for device in login_devices:
            if device['is_current'] == 1:
                device_flag = device['ssoent']
                if device_flag.startswith('A'):
                    return 'web'
                elif device_flag.startswith('P'):
                    return 'pc'
                else:
                    break
        return 'unknown'



__author__ = 'deadblue'

import os
import os.path
import typing

from py115._internal.api import offline, file, dir, space, upload, video
from py115._internal.protocol.client import Client

from py115 import types


_CLEAR_FLAG_MAPPING = {
    types.TaskStatus.Complete: 1,
    types.TaskStatus.Failed: 2,
    types.TaskStatus.Running: 3
}


class OfflineService:
    """Offline task manager."""

    _client: Client = None
    _app_ver: str = None
    _user_id: str = None

    @classmethod
    def _create(cls, client: Client, app_ver: str, user_id: str):
        s = cls()
        s._client = client
        s._app_ver = app_ver
        s._user_id = user_id
        return s

    def list(self) -> typing.Generator[types.Task, None, None]:
        """Get all tasks.

        Yields:
            py115.types.Task: Task object
        """
        spec = offline.ListApi()
        while True:
            result = self._client.execute_api(spec)
            for t in result['tasks']:
                yield types.Task._create(t)
            page, page_count = result['page'], result['page_count']
            if page < page_count:
                spec.set_page(page + 1)
            else:
                break

    def add_url(self, *urls: str) -> typing.Iterable[types.Task]:
        """Create task(s) from download URL.

        Args:
            *urls (str): Download URL, can be a http/ftp/ed2k/magnet link.

        Return:
            Iterable[py115.types.Task]: Task list for the download URLs.
        """
        if len(urls) == 0:
            return []
        add_results = self._client.execute_api(offline.AddUrlsApi(
            self._app_ver, self._user_id, urls
        ))
        return [types.Task._create(r) for r in add_results]

    def delete(self, *task_ids: str):
        """Delete task(s).

        Args:
            *task_ids (str): The ID of tasks you wants to delete.
        """
        if len(task_ids) == 0:
            return
        self._client.execute_api(offline.DeleteApi(task_ids))

    def clear(self, status: types.TaskStatus = types.TaskStatus.Complete):
        """Clear tasks.

        Args:
            status (py115.types.TaskStatus): 
                Tasks in given status will be cleared. Set status to None to 
                clear all tasks.
        """
        if status is None:
            # Clear all tasks
            flag = 1
        else:
            # Default to clear complete tasks.
            flag = _CLEAR_FLAG_MAPPING.get(status, 0)
        self._client.execute_api(offline.ClearApi(flag))


class StorageService:
    """Cloud file/directory manager."""

    _client: Client
    _app_type: str
    _upload_helper: upload.Helper

    @classmethod
    def _create(cls, client: Client, app_type: str, uh: upload.Helper):
        s = cls()
        s._client = client
        s._app_type = app_type
        s._upload_helper = uh
        return s

    def space(self) -> typing.Tuple[int, int]:
        """
        Get total size and used size of the storage.

        Returns:
            Tuple[int, int]: Total size and used size in byte.
        """
        result = self._client.execute_api(space.GetApi())
        if result is not None:
            total = int(result['all_total']['size'])
            used = int(result['all_use']['size'])
            return (total, used)
        else:
            return (0, 0)

    def list(self, dir_id: str = '0') -> typing.Generator[types.File, None, None]:
        """Get files under a directory.

        Args:
            dir_id (str): Directory ID to list, default is "0" which means root directory.

        Yields:
            py115.types.File: File object under the directory.
        """
        spec = file.ListApi(dir_id)
        while True:
            result = self._client.execute_api(spec)
            for f in result['files']:
                yield types.File._create(f)
            next_offset = result['offset'] + len(result['files'])
            if next_offset >= result['count']:
                break
            else:
                spec.set_offset(next_offset)

    def search(self, keyword: str, dir_id: str = '0'):
        """Recursively search files under a directory.

        Args:
            keyword (str): Keyword to search files.
            dir_id (str): Directory ID to search.

        Yields:
            py115.types.File: File object whose name contains the keyword.
        """
        spec = file.SearchApi(keyword=keyword, dir_id=dir_id)
        while True:
            result = self._client.execute_api(spec)
            for f in result['files']:
                yield types.File._create(f)
            next_offset = result['offset'] + len(result['files'])
            if next_offset >= result['count']:
                break
            else:
                spec.set_offset(next_offset)

    def move(self, target_dir_id: str, *file_ids: str):
        """Move files to a directory.

        Args:
            target_dir_id (str): ID of target directory where to move files.
            *file_ids (str): ID of files to be moved.
        """
        if len(file_ids) == 0:
            return
        self._client.execute_api(file.MoveApi(target_dir_id, file_ids))

    def rename(self, file_id: str, new_name: str):
        """Rename file.
        
        Args:
            file_id (str): ID of file to be renamed.
            new_name (str): New name for the file.
        """
        spec = file.RenameApi()
        spec.add_file(file_id, new_name)
        self._client.execute_api(spec)

    def batch_rename(self, *pairs: typing.Tuple[str, str]):
        """Batch rename files

        Args:
            *pair (Tuple[str, str]): Pair of file_id and new_name.
        """
        spec = file.RenameApi()
        for file_id, new_name in pairs:
            spec.add_file(file_id, new_name)
        self._client.execute_api(spec)

    def delete(self, *file_ids: str):
        """Delete files.

        Args:
            *file_ids (str): ID of files to be deleted.
        """
        if len(file_ids) == 0:
            return
        self._client.execute_api(file.DeleteApi(file_ids))

    def make_dir(self, parent_id: str, name: str) -> types.File:
        """Make new directory under a directory.
        
        Args:
            parent_id (str): ID of parent directory where to make new directory.
            name (str): Name for the new directory.
        
        Return:
            py115.types.File: File object of the created directory.
        """
        result = self._client.execute_api(dir.AddApi(parent_id, name))
        result['pid'] = parent_id
        return types.File._create(result)

    def request_download(self, pickcode: str) -> types.DownloadTicket:
        """Download file from cloud storage.

        Args:
            pickcode (str): Pick code of file.

        Returns:
            py115.types.DownloadTicket: A ticket contains all required fields to 
            download file from cloud.
        """
        result = self._client.execute_api(file.DownloadApi(pickcode))
        if result is None or 'url' not in result:
            return None
        ticket = types.DownloadTicket(
            url=result.get('url'),
            file_name=result.get('file_name'),
            file_size=result.get('file_size')
        )
        # Required headers for downloading
        cookies = self._client.export_cookies(url=result['url'])
        ticket.headers.update({
            'User-Agent': self._client.user_agent,
            'Cookie': '; '.join([
                f'{k}={v}' for k, v in cookies.items()
            ])
        })
        return ticket

    def request_upload(self, dir_id: str, file_path: str) -> types.UploadTicket:
        """Upload local file to cloud storage.

        Args:
            dir_id (str): ID of directory where to store the file.
            file_path (str): Path of the local file.
        
        Return:
            py115.types.UploadTicket: A ticket contains all required fields to
            upload file to cloud, should be used with aliyun-oss-python-sdk.
        """
        if not os.path.exists(file_path):
            return None
        file_name = os.path.basename(file_path)
        with open(file_path, 'rb') as file_io:
            return self.request_upload_data(dir_id, file_name, file_io)

    def request_upload_data(
            self, 
            dir_id: str, 
            save_name: str, 
            data_io: typing.BinaryIO, 
        ) -> types.UploadTicket:
        """Upload data as a file to cloud storage.

        Args:
            dir_id (str): ID of directory where to store the file.
            save_name (str): File name to be saved.
            data_io (BinaryIO): IO stream of data.

        Return:
            py115.types.UploadTicket: A ticket contains all required fields to
            upload file to cloud, should be used with aliyun-oss-python-sdk.
        """
        if not (data_io.readable() and data_io.seekable()):
            return None
        init_result = self._client.execute_api(upload.InitApi(
            target_id=f'U_1_{dir_id}',
            file_name=save_name,
            file_io=data_io,
            helper=self._upload_helper
        ))
        token_result = None
        if not init_result['done']:
            token_result = self._client.execute_api(upload.TokenApi())
        return types.UploadTicket._create(init_result, token_result)

    def request_play(self, pickcode: str) -> types.PlayTicket:
        """Play a video file on cloud, returns required parameters as a ticket.

        Args:
            pickcode (str): Pick code of file.
        
        Returns:
            py115.types.PlayTicket: A ticket contains all required fields to 
            play the media file on cloud.
        """
        if self._app_type == 'web':
            spec = video.WebPlayApi(pickcode=pickcode)
        else:
            spec = video.PcPlayApi(
                pickcode=pickcode,
                user_id=self._upload_helper.user_id,
                app_ver=self._upload_helper.app_version
            )
        play_url = self._client.execute_api(spec)
        ticket = types.PlayTicket(
            url=play_url
        )
        if self._app_type == 'web':
            # Prepare headers for playing
            cookies = self._client.export_cookies(url=ticket.url)
            ticket.headers.update({
                'User-Agent': self._client.user_agent,
                'Cookie': '; '.join([
                    f'{k}={v}' for k, v in cookies.items()
                ])
            })
        return ticket



__author__ = 'deadblue'

from typing import Any, Dict, Union

from py115._internal.protocol.client import Client
from py115._internal.api import app, qrcode, upload

from py115 import services
from py115.types import Credential, LoginTarget, QrcodeSession, QrcodeStatus


_app_names = [
    'web', 'mac', 'linux', 'windows'
]


class Cloud:
    """115 cloud service.

    Args:
        credential (Any): Credential object.
        protocol_kwargs (dict): Settings for underlying protocol client.
    """

    _client: Client = None
    _app_ver: str = None
    _app_type: str = None
    _upload_helper: upload.Helper = None

    def __init__(
            self, 
            credential: Any = None, 
            protocol_kwargs: dict = None
        ) -> None:
        # Config protocol client
        protocol_kwargs = protocol_kwargs or {}
        self._client = Client(**protocol_kwargs)
        # Get latest app version
        self._app_ver = self._client.execute_api(app.GetVersionApi())
        self._client.setup_user_agent(
            app_version=self._app_ver
        )
        # Set credential
        if credential is not None:
            self.import_credential(credential)

    def import_credential(self, credential: Any) -> bool:
        """Import credential to cloud instance.

        Args:
            credential (Any): Credential object to identity user.

        Return:
            bool: Is credential valid.
        """
        cookies = _extract_cookies(credential)
        if cookies is None:
            return False
        self._client.import_cookies(cookies)
        return self._check_login()

    def export_credentail(self) -> Union[Credential, None]:
        """Export current credentail from cloud instance.

        Return:
            py115.types.Credential: Credential object, or None when credential is invalid.
        """
        # TODO: Check credential before return.
        return Credential.from_dict(
            self._client.export_cookies()
        )

    def qrcode_login(self, target: LoginTarget = LoginTarget.Web) -> QrcodeSession:
        """Start QRcode login session.

        Args:
            app_type (py115.types.AppType): App to login.

        Returns:
            py115.types.QrcodeSession: QRcode login session.
        """
        app_name = _app_names[target.value]
        token = self._client.execute_api(qrcode.TokenApi(app_name))
        qrcode_image = self._client.fetch(
            qrcode.get_image_url(app_name, token.get('uid'))
        )
        return QrcodeSession._create(
            app_name=app_name,
            uid=token.get('uid'), 
            time=token.get('time'),
            sign=token.get('sign'),
            image=qrcode_image
        )

    def qrcode_poll(self, session: QrcodeSession) -> QrcodeStatus:
        """Poll QRcode login status.

        Args:
            session (py115.types.QrcodeSession): QRcode login session.
        
        Returns:
            py115.types.QrcodeStatus: QRcode status.
        """
        status = self._client.execute_api(qrcode.StatusApi(
            uid=session._uid, time=session._time, sign=session._sign
        ))
        if status == 2:
            self._client.execute_api(qrcode.LoginApi(
                app_name=session._app_name, 
                uid=session._uid
            ))
            if self._check_login():
                return QrcodeStatus.Done
            else:
                return QrcodeStatus.Failed
        elif status < 0:
            return QrcodeStatus.Expired
        return QrcodeStatus.Waiting

    def _check_login(self) -> bool:
        try:
            # Get login app type
            self._app_type = self._client.execute_api(app.GetTypeApi())
            # Get upload key
            user_id, user_key = self._client.execute_api(upload.InfoApi())
            self._upload_helper = upload.Helper(
                self._app_ver, user_id, user_key
            )
            return True
        except:
            return False

    def offline(self) -> services.OfflineService:
        """Get offline service.

        Return:
            py115.services.OfflineService: Offline service instance.
        """
        return services.OfflineService._create(
            client=self._client, 
            app_ver=self._app_ver, 
            user_id=self._upload_helper.user_id
        )

    def storage(self) -> services.StorageService:
        """Get storage service.

        Return:
            py115.services.StorageService: Storage service instance.
        """
        return services.StorageService._create(
            client=self._client,
            app_type=self._app_type,
            uh=self._upload_helper
        )

def _extract_cookies(credential: Any) -> Union[Dict[str, str], None]:
    if isinstance(credential, Credential):
        return credential.to_dict()
    # Handle dict or object
    raw: Dict[str, Any] = None
    if isinstance(credential, dict):
        raw = credential
    elif hasattr(credential, '__dict__'):
        raw = credential.__dict__
    if raw is None:
        return None
    # Extract cookies from credentials
    cookies = {}
    for key, value in raw.items():
        key = key.upper()
        if key in ('UID', 'CID', 'SEID'):
            cookies[key] = str(value)
    return cookies



__author__ = 'deadblue'

import binascii
import logging
import random
import struct

from Crypto.Cipher import AES
from Crypto.PublicKey import ECC
import lz4.block

_server_pub_key = bytes([
    0x04, 0x57, 0xa2, 0x92, 0x57, 0xcd, 0x23, 0x20, 
    0xe5, 0xd6, 0xd1, 0x43, 0x32, 0x2f, 0xa4, 0xbb, 
    0x8a, 0x3c, 0xf9, 0xd3, 0xcc, 0x62, 0x3e, 0xf5, 
    0xed, 0xac, 0x62, 0xb7, 0x67, 0x8a, 0x89, 0xc9, 
    0x1a, 0x83, 0xba, 0x80, 0x0d, 0x61, 0x29, 0xf5, 
    0x22, 0xd0, 0x34, 0xc8, 0x95, 0xdd, 0x24, 0x65, 
    0x24, 0x3a, 0xdd, 0xc2, 0x50, 0x95, 0x3b, 0xee, 
    0xba,
])

_curve_name = 'P-224'

_crc_salt = b'^j>WD3Kr?J2gLFjD4W2y@'

_logger = logging.getLogger(__name__)

class Cipher:

    def __init__(self) -> None:
        # Load server public key
        server_key = ECC.import_key(
            encoded=_server_pub_key, curve_name=_curve_name
        )
        # Generate client key
        client_key = ECC.generate(curve=_curve_name)
        # Export client public key
        self._pub_key = b'\x1d' + client_key.public_key().export_key(
            format='SEC1', compress=True
        )
        # ECDH key exchange
        shared_secret = (server_key.pointQ * client_key.d).x.to_bytes(28, 'big')
        self._aes_key = shared_secret[:16]
        self._aes_iv = shared_secret[-16:]
    
    def encode_token(self, timestamp: int) -> str:
        token = bytearray(struct.pack(
            '<15sBII15sBI',
            self._pub_key[:15], 0, 115, timestamp,
            self._pub_key[15:], 0, 1
        ))
        r1, r2 = random.randint(0, 0xff), random.randint(0, 0xff)
        for i in range(len(token)):
            if i < 24:
                token[i] = token[i] ^ r1
            else:
                token[i] = token[i] ^ r2
        # Calculate and append CRC32 checksum
        checksum = binascii.crc32(_crc_salt + token) & 0xffffffff
        token += struct.pack('<I', checksum)
        # Base64 encode
        return binascii.b2a_base64(token, newline=False).decode()

    def encode(self, data: bytes) -> bytes:
        pad_size = AES.block_size - len(data) % AES.block_size
        if pad_size != AES.block_size:
            data += b'\x00' * pad_size
        encrypter = AES.new(
            key=self._aes_key,
            mode=AES.MODE_CBC,
            iv=self._aes_iv
        )
        return encrypter.encrypt(data)

    def decode(self, data: bytes) -> bytes:
        ciphertext, tail = data[:-12], bytearray(data[-12:])
        # Decrypt
        decrypter = AES.new(
            key = self._aes_key, 
            mode = AES.MODE_CBC, 
            iv = self._aes_iv
        )
        plaintext = decrypter.decrypt(ciphertext)
        # Get uncompress size
        for i in range(4):
            tail[i] = tail[i] ^ tail[7]
        dst_size, = struct.unpack('<I', tail[:4])
        # Decompress
        buf = []
        while dst_size > 0:
            uncompressed_size = dst_size
            if uncompressed_size > 8192:
                uncompressed_size = 8192
            src_size, = struct.unpack('<H', plaintext[:2])
            buf.append(lz4.block.decompress(
                plaintext[2:src_size+2], uncompressed_size
            ))
            # Move to next block
            plaintext = plaintext[src_size+2:]
            dst_size -= uncompressed_size
        return buf[0] if len(buf) == 0 else b''.join(buf)



__author__ = 'deadblue'

import json
import logging
import time
import warnings
from urllib.parse import urlparse
from urllib3.exceptions import InsecureRequestWarning

import requests

from py115._internal.crypto import ec115
from py115._internal.protocol.api import ApiSpec, RetryException


_logger = logging.getLogger(__name__)


class Client:

    def __init__(self, **kwargs) -> None:
        self._ecc = ec115.Cipher()
        self._session = requests.Session()
        # Configure session
        self._user_agent = 'Mozilla/5.0 115Browser/27.0.3.7'
        self._session.headers.update({
            'User-Agent': self._user_agent
        })
        # Flow control
        self._next_request_time = 0.0
        # Protocol client settings
        proxy_addr = kwargs.pop('proxy', None)
        if proxy_addr is not None and isinstance(proxy_addr, str):
            self._session.proxies = {
                'http': proxy_addr,
                'https': proxy_addr
            }
        verify = kwargs.pop('verify', None)
        if verify is not None and isinstance(verify, bool):
            self._session.verify = verify
            if not verify:
                warnings.simplefilter('ignore', category=InsecureRequestWarning)

    def import_cookies(self, cookies: dict):
        for name, value in cookies.items():
            self._session.cookies.set(
                name, value,
                domain='.115.com',
                path='/'
            )

    def export_cookies(self, url: str = None) -> dict:
        req_domain, req_path = '.115.com', '/'
        if url is not None:
            result = urlparse(url)
            req_domain = f'.{result.hostname}'
            req_path = f'{result.path}/'

        cookie_dict = {}
        for cookie in self._session.cookies:
            # Check domain
            cookie_domain = cookie.domain
            if not cookie_domain.startswith('.'):
                cookie_domain = f'.{cookie_domain}'
            if not req_domain.endswith(cookie_domain): continue
            # Check path
            cookie_path = cookie.path
            if not cookie_path.endswith('/'):
                cookie_path = f'{cookie_path}/'
            if not req_path.startswith(cookie_path): continue
            # Add cookie to dict
            cookie_dict[cookie.name] = cookie.value
        return cookie_dict

    def setup_user_agent(self, app_version: str):
        self._user_agent = 'Mozilla/5.0 115Browser/%s' % app_version
        self._session.headers.update({
            'User-Agent': self._user_agent
        })

    @property
    def user_agent(self) -> str:
        return self._user_agent

    def execute_api(self, spec: ApiSpec):
        while True:
            try:
                return self._execute_api_internal(spec)
            except RetryException:
                pass

    def _execute_api_internal(self, spec: ApiSpec):
        if spec.use_ec:
            spec.update_qs({
                'k_ec': self._ecc.encode_token(int(time.time()))
            })
        data = spec.payload
        # Prepare request
        req = requests.Request(
            method='GET',
            url=spec.url,
            params=spec.qs
        )
        if data is not None:
            if spec.use_ec:
                data = self._ecc.encode(data)
            req.method = 'POST'
            req.data = data
            req.headers.update({
                'Content-Type': 'application/x-www-form-urlencoded'
            })
        # Send request with retrying
        while True:
            # Flow control
            wait_time = self._next_request_time - time.time()
            if wait_time > 0:
                time.sleep(wait_time)
            try:
                resp = self._session.send(
                    request=self._session.prepare_request(req),
                    timeout=(10.0, 30.0)
                )
                break
            except requests.Timeout:
                _logger.warning('Request timeout, retry ...')
            finally:
                self._next_request_time = time.time() + spec.get_delay()
        # Decrypt response body
        if spec.use_ec:
            result = json.loads(self._ecc.decode(resp.content))
        else:
            result = resp.json()
        _logger.debug('API result: %r', result)
        return spec.parse_result(result)

    def fetch(self, url: str) -> bytes:
        resp = self._session.get(url)
        return resp.content



__author__ = 'deadblue'

import time
import typing

from py115._internal.api.m115 import M115ApiSpec
from py115._internal.protocol import api


class ListApi(api.ApiSpec):

    def __init__(self, dir_id: str) -> None:
        super().__init__(None)
        self.update_qs({
            'aid': '1',
            'cid': dir_id,
            'o': 'user_ptime',
            'asc': '0',
            'offset': '0',
            'limit': '115',
            'show_dir': '1',
            'snap': '0',
            'natsort': '1',
            'format': 'json',
        })

    @property
    def url(self) -> str:
        order = self._qs.get('o')
        if order == 'file_name':
            return 'https://aps.115.com/natsort/files.php'
        else:
            return 'https://webapi.115.com/files'

    def parse_result(self, result: dict) -> dict:
        if result.get('state', False):
            return {
                'count': result.get('count'),
                'offset': result.get('offset'),
                'files': result.get('data')
            }
        else:
            err_code = api.find_error_code(result)
            if err_code == 20130827:
                self._qs.update({
                    'o': result['order'],
                    'asc': result['is_asc'],
                })
                raise api.RetryException()
            else:
                raise api.ApiException(err_code)

    def set_offset(self, offset: int):
        self.update_qs({
            'offset': str(offset)
        })


class SearchApi(api.ApiSpec):

    def __init__(self, keyword: str, dir_id: str) -> None:
        super().__init__('https://webapi.115.com/files/search')
        self.update_qs({
            'aid': '1',
            'cid': dir_id,
            'offset': '0',
            'limit': '115',
            'format': 'json',
            'search_value': keyword
        })
    
    def parse_result(self, result: dict) -> dict:
        if result.get('state', False):
            return {
                'count': result.get('count'),
                'offset': result.get('offset'),
                'files': result.get('data')
            }
        else:
            raise api.ApiException(api.find_error_code(result))

    def set_offset(self, offset: int):
        self.update_qs({
            'offset': str(offset)
        })


class DeleteApi(api.ApiSpec):

    def __init__(self, file_ids: typing.Iterable[str]) -> None:
        super().__init__('https://webapi.115.com/rb/delete')
        self.update_from({
            'fid': file_ids,
            'ignore_warn': '1'
        })


class MoveApi(api.ApiSpec):

    def __init__(self, parent_id: str, file_ids: typing.Iterable[str]) -> None:
        super().__init__('https://webapi.115.com/files/move')
        self.update_from({
            'pid': parent_id,
            'fid': file_ids
        })


class RenameApi(api.ApiSpec):

    def __init__(self) -> None:
        super().__init__('https://webapi.115.com/files/batch_rename')

    def add_file(self, file_id: str, new_name: str):
        key = 'files_new_name[%s]' % file_id
        self.update_from({
            key: new_name
        })


class DownloadApi(M115ApiSpec):

    def __init__(self, pickcode: str) -> None:
        super().__init__('https://proapi.115.com/app/chrome/downurl', True)
        self.update_qs({
            't': int(time.time())
        })
        self.update_from({
            'pickcode': pickcode
        })

    def parse_result(self, result: dict):
        result = super().parse_result(result)
        if len(result) == 0:
            return None
        file_id, down_info = result.popitem()
        if 'url' in down_info and isinstance(down_info['url'], dict):
            return {
                'file_id': file_id,
                'file_name': down_info['file_name'],
                'file_size': int(down_info['file_size']),
                'url': down_info['url']['url'].replace('http://', 'https://')
            }
        else:
            return None



__author__ = 'deadblue'

import typing

from py115._internal.api import m115
from py115._internal.protocol import api


class ListApi(api.ApiSpec):
    
    def __init__(self, page: int = 1) -> None:
        super().__init__('https://lixian.115.com/lixian/', True)
        self.update_qs({
            'ct': 'lixian',
            'ac': 'task_lists',
            'page': page
        })

    def set_page(self, page: int):
        self.update_qs({
            'page': page
        })

    def parse_result(self, result: dict) -> typing.Any:
        error_code = api.find_error_code(result)
        if error_code != 0:
            raise api.ApiException(error_code)
        tasks = result.get('tasks', None)
        return {
            'page_count': result.get('page_count', 1),
            'page': result.get('page', 1),
            'task_count': result.get('count', 0),
            'tasks': tasks or []
        }


class DeleteApi(api.ApiSpec):

    def __init__(self, task_ids: typing.Iterable[str]) -> None:
        super().__init__('https://lixian.115.com/lixian/', True)
        self.update_qs({
            'ct': 'lixian',
            'ac': 'task_del'
        })
        self.update_from({
            'hash': task_ids,
            'flag': '0',
        })


class ClearApi(api.ApiSpec):

    def __init__(self, flag: int) -> None:
        super().__init__('https://lixian.115.com/lixian/', True)
        self.update_qs({
            'ct': 'lixian',
            'ac': 'task_clear'
        })
        self.update_from({
            'flag': flag
        })


class AddUrlsApi(m115.M115ApiSpec):
    
    def __init__(self, app_ver: str, user_id: int, urls: typing.Iterable[str], **kwargs) -> None:
        super().__init__('https://lixian.115.com/lixianssp/', True)
        self.update_qs({
            'ac': 'add_task_urls'
        })
        params = {
            'ac': 'add_task_urls',
            'app_ver': app_ver,
            'uid': user_id,
            'url': urls
        }
        save_dir_id = kwargs.pop('save_dir_id', None)
        if save_dir_id is not None:
            params['wp_path_id'] = save_dir_id
        self.update_from(params)
    
    def parse_result(self, result: dict):
        result = super().parse_result(result)
        return result['result']



__author__ = 'deadblue'

import time as timelib
import typing

from py115._internal.protocol import api


_app_id_mapping = {
    'web': 0,
    'mac': 7,
    'linux': 7,
    'windows': 7,
}


class _BaseApi(api.ApiSpec):

    def parse_result(self, result: dict) -> typing.Any:
        if result.get('state', 0) != 1:
            raise api.ApiException(code=result.get('code'))
        return result.get('data')


class TokenApi(_BaseApi):

    def __init__(self, app_name: str) -> None:
        super().__init__(
            f'https://qrcodeapi.115.com/api/1.0/{app_name}/1.0/token', True
        )


class StatusApi(_BaseApi):

    def __init__(self, uid: str, time: int, sign: str) -> None:
        super().__init__('https://qrcodeapi.115.com/get/status/', False)
        self.update_qs({
            'uid': uid,
            'time': time,
            'sign': sign,
            '_': int(timelib.time())
        })

    def parse_result(self, result: dict) -> int:
        code = result.get('code', 0)
        if code == 40199002:
            return -9
        elif code != 0:
            raise api.ApiException(code)
        result_data = result.get('data', {})
        return result_data.get('status', 0)


class LoginApi(_BaseApi):

    def __init__(self, app_name: str, uid: str) -> None:
        super().__init__(
            f'https://passportapi.115.com/app/1.0/{app_name}/1.0/login/qrcode', True
        )
        self.update_from({
            'account': uid,
            'app': app_name
        })


def get_image_url(app_name: str, uid: str) -> str:
    app_id = _app_id_mapping.get(app_name, 0)
    return f'https://qrcodeapi.115.com/api/1.0/{app_name}/1.0/qrcode?qrfrom=1&client={app_id}d&uid={uid}'



__author__ = 'deadblue'

import secrets


class Cipher():

    def __init__(self, n: bytes, e: int) -> None:
        self._n = int.from_bytes(n, 'big')
        self._e = e
        self._key_len = len(n)


    def _encrypt_slice(self, segment: bytes) -> bytes:
        # Make pad
        pad_size = self._key_len - len(segment)
        pad_buf = bytearray(pad_size)
        pad_buf[0], pad_buf[1] = 0, 2
        for i in range(pad_size - 3):
            pad_buf[2 + i] = secrets.randbelow(0xff) + 1
        pad_buf[-1] = 0
        msg = int.from_bytes(pad_buf + segment, 'big')
        return pow(msg, self._e, self._n).to_bytes(self._key_len, 'big')


    def encrypt(self, plaintext: bytes) -> bytes:
        ciphertext = bytearray()
        remain_size = len(plaintext)
        while remain_size > 0:
            slice_size = self._key_len - 11
            if slice_size > remain_size:
                slice_size = remain_size
            ciphertext.extend(
                self._encrypt_slice(plaintext[:slice_size])
            )
            plaintext = plaintext[slice_size:]
            remain_size -= slice_size
        return bytes(ciphertext)


    def _decrypt_slice(self, segment: bytes) -> bytes:
        msg = int.from_bytes(segment, 'big')
        ret = pow(msg, self._e, self._n).to_bytes(self._key_len, 'big')
        for i, b in enumerate(ret):
            if i != 0 and b == 0:
                ret = ret[i+1:]
                break
        return ret


    def decrypt(self, ciphertext: bytes) -> bytearray:
        plaintext = bytearray()
        remain_size = len(ciphertext)
        while remain_size > 0:
            slice_size = self._key_len
            if slice_size > remain_size:
                slice_size = remain_size
            plaintext.extend(
                self._decrypt_slice(ciphertext[:slice_size])
            )
            ciphertext = ciphertext[slice_size:]
            remain_size -= slice_size
        return plaintext



__author__ = 'deadblue'

import datetime

import pytz

_tz_cst = pytz.FixedOffset(480)

def parse_datetime_str(s: str) -> datetime.datetime:
    if '-' in s:
        t = datetime.datetime.strptime(
            s, '%Y-%m-%d %H:%M'
        )
        t.replace(tzinfo=_tz_cst)
        return t
    else:
        return  datetime.datetime.fromtimestamp(
            float(s), tz=_tz_cst
        )

def make_datetime(ts: int) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(
        float(ts), tz=_tz_cst
    )



__author__ = 'deadblue'

from py115._internal.api import m115
from py115._internal.protocol import api


class WebPlayApi(api.ApiSpec):

    def __init__(self, pickcode: str) -> None:
        super().__init__('https://webapi.115.com/files/video', False)
        self.update_qs({'pickcode': pickcode})
    
    def parse_result(self, result: dict) -> str:
        err_code = api.find_error_code(result)
        if err_code != 0:
            raise api.ApiException(code=err_code)
        return result.get('video_url')


class PcPlayApi(m115.M115ApiSpec):

    def __init__(self, pickcode: str, user_id: str, app_ver: str) -> None:
        super().__init__('https://proapi.115.com/pc/video/play')
        self.update_from({
            'format': 'app',
            'definition_filter': '1',
            'pickcode': pickcode,
            'user_id': user_id,
            'appversion': app_ver
        })
    
    def parse_result(self, result: dict):
        result = super().parse_result(result)
        video_urls = sorted(
            result['video_url'], 
            key=lambda u: u['definition'],
            reverse=True
        )
        return video_urls[0]['url']



__author__ = 'deadblue'

from py115._internal.protocol import api

class GetApi(api.ApiSpec):

    def __init__(self) -> None:
        super().__init__('https://webapi.115.com/files/index_info')
    
    def parse_result(self, result: dict):
        data = super().parse_result(result)
        return data.get('space_info', None)



__author__ = 'deadblue'

from datetime import datetime
from enum import IntEnum
from typing import Self, Optional

from py115._internal import oss, utils


class LoginTarget(IntEnum):
    """App to login."""

    Web = 0
    """Login as web"""
    Mac = 1
    """Login as MAC app"""
    Linux = 2
    """Login as Linux app"""
    Windows = 3
    """Login as Windows app"""


class QrcodeStatus(IntEnum):

    Waiting = 0
    """QRCode is waiting for scanning."""

    Done = 1
    """QRCode is scanned and approved by user."""

    Expired = -1
    """QRCode is expired."""

    Failed = -2
    """QRcode is declined by user."""


class QrcodeSession:
    """QRcode login session."""

    _app_name: str
    _uid: str
    _time: int
    _sign: str

    image_data: bytes
    """QRcode image data"""

    @classmethod
    def _create(cls, app_name: str, uid: str, time: int, sign: str, image: bytes) -> Self:
        ret = cls()
        ret._app_name = app_name
        ret._uid = uid
        ret._time = time
        ret._sign = sign
        ret.image_data = image
        return ret


class _Base:
    """Base clase for all types."""

    def __repr__(self) -> str:
        cls_name = type(self).__name__
        attrs = ', '.join([
            f'{k}={v}' for k, v in self.__dict__.items()
        ])
        return f'{cls_name}({attrs})'


class Credential(_Base):
    """Credential contains required information to identify user.

    Args:
        uid (str): The "UID" value in cookies.
        cid (str): The "CID" value in cookies.
        seid (str): The "SEID" value in cookies.
    """

    _uid: str = None
    _cid: str = None
    _seid: str = None

    def __init__(self, uid: str = None, cid: str = None, seid: str = None) -> None:
        self._uid = uid
        self._cid = cid
        self._seid = seid

    @classmethod
    def from_dict(cls, d: dict):
        """Create Credential from a dict object.

        Args:
            d (dict): Dict object.

        Returns:
            py115.types.Credential: Credential object.
        """
        if len(d) == 0 or not ('UID' in d and 'CID' in d and 'SEID' in d):
            return None
        return Credential(
            uid=d.get('UID'),
            cid=d.get('CID'),
            seid=d.get('SEID')
        )

    def to_dict(self) -> dict:
        """Convert credential object to dict object.

        Returns:
            dict: Dict object.
        """
        return {
            'UID': self._uid,
            'CID': self._cid,
            'SEID': self._seid,
        }


class File(_Base):
    """File represents a file or directory on cloud storage."""

    file_id: str
    """Unique ID of the file on cloud."""

    parent_id: str
    """File ID of the parent directory."""

    name: str
    """Base name."""

    modified_time: datetime
    """Last modified datetime."""

    is_dir: bool
    """Is file a directory."""

    size: int = 0
    """Size in bytes."""

    sha1: Optional[str] = None
    """SHA-1 hash of the file in HEX encoding."""

    pickcode: Optional[str] = None
    """Pickcode to download the file."""

    media_duration: Optional[int] = None
    """Media duration in seconds for audio/video file."""


    @classmethod
    def _create(cls, raw: dict) -> Self:
        category_id = raw.get('cid')
        file_id = raw.get('fid', None)
        parent_id = raw.get('pid', None)

        r = cls()
        r.name = raw.get('n')
        r.size = raw.get('s', 0)
        r.sha1 = raw.get('sha', None)
        r.media_duration = raw.get('play_long', None)
        if 'te' in raw:
            r.modified_time = utils.parse_datetime_str(raw.get('te'))
        elif 't' in raw:
            r.modified_time = utils.parse_datetime_str(raw.get('t'))
        else:
            r.modified_time = datetime.now()
        r.is_dir = file_id is None
        if r.is_dir:
            r.file_id = category_id
            r.parent_id = parent_id
            r.pickcode = None
        else:
            r.file_id = file_id
            r.parent_id = category_id
            r.pickcode = raw.get('pc')
        return r

    def __str__(self) -> str:
        if self.is_dir:
            return f'{self.name}/'
        else:
            return self.name


class DownloadTicket(_Base):
    """
    DownloadTicket contains required parameters to download a file from 
    cloud storage.

    Please check examples for detail usage.
    """

    file_name: str
    """Base name of the file."""

    file_size: int
    """Size of the file."""

    url: str
    """Download URL."""

    headers: dict
    """Required headers that should be used with download URL."""

    def __init__(self, url: str, file_name: str, file_size: int) -> None:
        self.url = url
        self.file_name = file_name
        self.file_size = file_size
        self.headers = {}


class PlayTicket(_Base):

    url: str
    """Download URL."""

    headers: dict
    """Required headers that should be used with download URL."""

    def __init__(self, url: str) -> None:
        self.url = url
        self.headers = {}


class UploadTicket(_Base):
    """
    UploadTicket contains required parameters to upload a file to 
    cloud storage.

    Please check examples for detial usage.
    """

    is_done: bool
    """Is file already uploaded."""

    oss_endpoint: str
    """OSS endpoint address."""

    oss_token: dict
    """OSS token"""

    bucket_name: str
    """OSS bucket name."""

    object_key: str
    """OSS object key."""

    headers: dict
    """Required headers that should be used in upload."""

    expiration: datetime
    """Expiration time of this ticket."""

    @classmethod
    def _create(cls, raw: dict, token: dict) -> Self:
        r = cls()
        r.is_done = raw['done']
        if not r.is_done:
            r.oss_endpoint = oss.ENDPOINT
            r.oss_token = {
                'access_key_id': token.get('access_key_id'),
                'access_key_secret': token.get('access_key_secret'),
                'security_token': token.get('security_token')
            }
            r.bucket_name = raw.get('bucket')
            r.object_key = raw.get('object')
            r.headers = {
                'x-oss-callback': oss.encode_header_value(raw.get('callback')),
                'x-oss-callback-var': oss.encode_header_value(raw.get('callback_var'))
            }
            r.expiration = token.get('expiration')
        return r
    
    def is_valid(self) -> bool:
        """Check whether the ticket is valid.

        Returns:
            bool: Valid flag.
        """
        return datetime.now() < self.expiration


class TaskStatus(IntEnum):
    Running = 1
    """Task is running."""
    Complete = 2
    """Task is complete."""
    Failed = -1
    """Task is failed."""
    Unknown = 0
    """Unknown status?"""


class Task(_Base):
    """Task represents an offline task."""

    task_id: str
    """Unique ID of the task."""

    name: str
    """Task name."""

    size: int
    """Total size to be downloaded."""

    created_time: datetime
    """Task created time."""

    percent: float
    """Precentage of the download, 0~100."""

    file_id: str
    """Downloaded file ID of the task, may be None if the task does not finish."""

    file_is_dir: bool
    """Is the downloaded file a directory."""

    status: TaskStatus
    """Task status."""

    @classmethod
    def _create(cls, raw: dict):
        r = cls()
        r.task_id = raw.get('info_hash')
        r.name = raw.get('name')
        r.size = raw.get('size', -1)
        r.created_time = utils.make_datetime(raw.get('add_time', 0))
        r.precent = float(raw.get('percentDone', 0))
        r.status = TaskStatus(raw.get('status', 0))

        file_id = raw.get('file_id', '')
        del_file_id = raw.get('delete_file_id', '')
        r.file_id = del_file_id
        r.file_is_dir = file_id != '' and file_id == del_file_id
        return r

    def is_complete(self) -> bool:
        """Check is task complete."""
        return self.status == TaskStatus.Complete

    def is_failed(self) -> bool:
        """Check is task failed."""
        return self.status == TaskStatus.Failed

    def is_running(self) -> bool:
        """Check is task still running."""
        return self.status == TaskStatus.Running


__author__ = 'deadblue'

from py115._internal.protocol import api


class AddApi(api.ApiSpec):

    def __init__(self, parent_id: str, dir_name: str) -> None:
        super().__init__('https://webapi.115.com/files/add')
        self.update_from({
            'pid': parent_id,
            'cname': dir_name
        })
    
    def parse_result(self, result: dict) -> dict:
        err_code = api.find_error_code(result)
        if err_code != 0:
            raise api.ApiException(err_code)
        return {
            'cid': result.get('cid'),
            'n': result.get('cname')
        }


class SortApi(api.ApiSpec):

    def __init__(self, dir_id: str, order: str, asc: bool = False) -> None:
        super().__init__('https://webapi.115.com/files/order')
        self.update_from({
            'file_id': dir_id,
            'user_order': order,
            'user_asc': '1' if asc else '0',
            'fc_mix': '0'
        })



__author__ = 'deadblue'

"""
Python SDK for 115 cloud storage service.

Example:

import py115
import py115.types

cloud = py115.connect(py115.types.Credential(
    uid='', cid='', seid=''
))
storage = cloud.storage()

for file in storage.list():
    print(f'File: {file.name}')

"""

VERSION = '0.0.6'

from typing import Any

from py115.cloud import Cloud

def connect(
        credential: Any = None,
        protocol_kwargs: dict=  None
) -> Cloud:
    """Connect to 115 cloud.

    Args:
        credential (Any): 
            Credential data to identity user.
        protocol_kwargs (dict): 
            Keyword arguments for underlying protocol client.

    Return:
        py115.cloud.Cloud: Cloud instance.
    """
    return Cloud(
        credential=credential, 
        protocol_kwargs=protocol_kwargs
    )



__author__ = 'deadblue'

import base64
import secrets

from py115._internal.crypto import _rsa, _xor


_rsa_n = bytes([
    0x86, 0x86, 0x98, 0x0c, 0x0f, 0x5a, 0x24, 0xc4, 
    0xb9, 0xd4, 0x30, 0x20, 0xcd, 0x2c, 0x22, 0x70, 
    0x3f, 0xf3, 0xf4, 0x50, 0x75, 0x65, 0x29, 0x05, 
    0x8b, 0x1c, 0xf8, 0x8f, 0x09, 0xb8, 0x60, 0x21, 
    0x36, 0x47, 0x71, 0x98, 0xa6, 0xe2, 0x68, 0x31, 
    0x49, 0x65, 0x9b, 0xd1, 0x22, 0xc3, 0x35, 0x92, 
    0xfd, 0xb5, 0xad, 0x47, 0x94, 0x4a, 0xd1, 0xea, 
    0x4d, 0x36, 0xc6, 0xb1, 0x72, 0xaa, 0xd6, 0x33, 
    0x8c, 0x3b, 0xb6, 0xac, 0x62, 0x27, 0x50, 0x2d, 
    0x01, 0x09, 0x93, 0xac, 0x96, 0x7d, 0x1a, 0xef, 
    0x00, 0xf0, 0xc8, 0xe0, 0x38, 0xde, 0x2e, 0x4d, 
    0x3b, 0xc2, 0xec, 0x36, 0x8a, 0xf2, 0xe9, 0xf1, 
    0x0a, 0x6f, 0x1e, 0xda, 0x4f, 0x72, 0x62, 0xf1, 
    0x36, 0x42, 0x0c, 0x07, 0xc3, 0x31, 0xb8, 0x71, 
    0xbf, 0x13, 0x9f, 0x74, 0xf3, 0x01, 0x0e, 0x3c, 
    0x4f, 0xe5, 0x7d, 0xf3, 0xaf, 0xb7, 0x16, 0x83, 
])

_rsa_e = 0x10001

_cipher = _rsa.Cipher(_rsa_n, _rsa_e)


def generate_key() -> bytes:
    return secrets.token_bytes(16)


def encode(key: bytes, input: str) -> str:
    input_data = bytearray(input.encode())
    _xor.transform(
        key=_xor.derive_key(key, 4),
        data=input_data
    )
    input_data.reverse()
    _xor.transform(
        key=_xor.client_key,
        data=input_data
    )
    # Prepend key
    input_data = key + input_data
    return base64.b64encode(
        _cipher.encrypt(input_data)
    ).decode()


def decode(key: bytes, input: str) -> str:
    input_data = bytearray( base64.b64decode(input) )
    plaintext = _cipher.decrypt(input_data)

    server_key, plaintext = plaintext[:16], plaintext[16:]
    _xor.transform(
        key=_xor.derive_key(server_key, 12),
        data=plaintext
    )
    plaintext.reverse()
    _xor.transform(
        key=_xor.derive_key(key, 4),
        data=plaintext
    )
    return plaintext.decode()



__author__ = 'deadblue'

import random
import typing
import urllib.parse


class ApiException(Exception):

    def __init__(self, code: int, *args: object) -> None:
        super().__init__(*args)
        self._code = code

    @property
    def error_code(self) -> int:
        return self._code


class RetryException(Exception):

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


_ERROR_CODE_FIELDS = [
    'errcode', 'errNo', 'errno', 'code'
]


def find_error_code(result: dict) -> int:
    if result.get('state', False):
        return 0
    for field in _ERROR_CODE_FIELDS:
        if field not in result:
            continue
        ec = result[field]
        if isinstance(ec, int) and ec > 0:
            return ec
    return -1


def _flat_params(params: dict) -> dict:
    flatted_params = {}
    for key, value in params.items():
        if isinstance(value, typing.Dict):
            for sub_k, sub_v in value.items():
                sub_k = '%s[%s]' % (key, sub_k)
                flatted_params[sub_k], str(sub_v)
        elif isinstance(value, (typing.List, typing.Tuple)):
            for sub_i, sub_v in enumerate(value):
                sub_k = '%s[%d]' % (key, sub_i)
                flatted_params[sub_k] = str(sub_v)
        else:
            flatted_params[key] = str(value)
    return flatted_params


class ApiSpec:

    def __init__(self, url: str, use_ec: bool = False) -> None:
        self._url = url
        self._use_ec = use_ec
        self._qs = {}
        self._form = {}

    @property
    def url(self) -> str:
        return self._url

    @property
    def use_ec(self) -> bool:
        return self._use_ec

    @property
    def qs(self) -> dict:
        return self._qs

    @property
    def payload(self) -> bytes:
        if len(self._form) > 0:
            return urllib.parse.urlencode(self._form).encode()
        else:
            return None

    def update_qs(self, params: dict):
        if self._qs is None:
            self._qs = _flat_params(params)
        else:
            self._qs.update(_flat_params(params))

    def update_from(self, params: dict):
        if self._form is None:
            self._form = _flat_params(params)
        else:
            self._form.update(_flat_params(params))

    def parse_result(self, result: dict) -> typing.Any:
        if 'data' in result:
            return result['data']
        error_code = find_error_code(result)
        if error_code == 0:
            return None
        else:
            raise ApiException(error_code)

    def get_delay(self) -> float:
        return random.randint(100, 500) / 1000.0



__author__ = 'deadblue'

import datetime
import hashlib
import io
import logging
import time
import typing

from py115._internal.protocol import api


_logger = logging.getLogger(__name__)

_token_salt = 'Qclm8MGWUv59TnrR0XPg'


class Helper:

    _app_ver: str = None
    _user_id: str = None
    _user_key: str = None
    _user_hash: str = None

    def __init__(self, app_ver: str, user_id: int, user_key: str) -> None:
        self._app_ver = app_ver
        self._user_id = str(user_id)
        self._user_key = user_key
        self._user_hash = hashlib.md5(
            self._user_id.encode()
        ).hexdigest()
    
    @property
    def app_version(self) -> str:
        return self._app_ver
    
    @property
    def user_id(self) -> str:
        return self._user_id

    def calc_sig(self, file_id: str, target_id: str) -> str:
        h1 = hashlib.sha1(
            f'{self._user_id}{file_id}{target_id}0'.encode()
        ).hexdigest().lower()
        return hashlib.sha1(
            f'{self._user_key}{h1}000000'.encode()
        ).hexdigest().upper()

    def calc_token(
            self, 
            file_id: str,
            file_size: int,
            timestamp: int,
            sign_key: str = '',
            sign_val: str = ''
        ) -> str:
        token_data = ''.join([
            _token_salt, file_id, str(file_size), sign_key, sign_val, 
            self._user_id, str(timestamp), self._user_hash, self._app_ver
        ]).encode()
        return hashlib.md5(token_data).hexdigest().lower()


class InfoApi(api.ApiSpec):

    def __init__(self) -> None:
        super().__init__('https://proapi.115.com/app/uploadinfo')

    def parse_result(self, result: dict) -> typing.Tuple[int, str]:
        err_code = api.find_error_code(result)
        if err_code != 0:
            raise api.ApiException(err_code)
        return result.get('user_id'), result.get('userkey')


class InitApi(api.ApiSpec):

    _helper: Helper = None
    _file_io: typing.BinaryIO = None

    def __init__(
            self, 
            target_id: str, 
            file_name: str, 
            file_io: typing.BinaryIO,
            helper: Helper
        ) -> None:
        super().__init__('https://uplb.115.com/4.0/initupload.php', True)
        self._file_io = file_io
        self._helper = helper
        now = int(time.time())
        file_size, file_hash = _digest_file(file_io)
        self.update_from({
            'appid': '0',
            'appversion': helper.app_version,
            'userid': helper.user_id,
            'filename': file_name,
            'filesize': file_size,
            'fileid': file_hash,
            'target': target_id,
            'sig': helper.calc_sig(file_hash, target_id),
            't': now,
            'token': helper.calc_token(file_hash, file_size, now)
        })

    def parse_result(self, result: dict):
        status = result['status']
        if status == 7:
            sign_key = result['sign_key']
            sign_val = _digest_file_range(self._file_io, result['sign_check'])
            now = int(time.time())
            self.update_from({
                'sign_key': sign_key,
                'sign_val': sign_val,
                't': now,
                'token': self._helper.calc_token(
                    self._form['fileid'], 
                    self._form['filesize'],
                    now, sign_key, sign_val
                )
            })
            raise api.RetryException()
        elif status == 2:
            return  {
                'done': True
            }
        elif status == 1:
            return {
                'done': False,
                'bucket': result['bucket'],
                'object': result['object'],
                'callback': result['callback']['callback'],
                'callback_var': result['callback']['callback_var']
            }
        else:
            _logger.warning('Unexpected upload init status: %d', status)
            return None


def _digest_file(r: typing.BinaryIO) -> typing.Tuple[int, str]:
    h = hashlib.sha1()
    while True:
        chunk = r.read(4096)
        if len(chunk) == 0: break
        h.update(chunk)
    file_size = r.tell()
    file_hash = h.hexdigest().upper()
    return file_size, file_hash


def _digest_file_range(r: typing.BinaryIO, check_range: str) :
    tmp = [int(x) for x in check_range.split('-')]
    start, length = tmp[0], tmp[1] - tmp[0] + 1
    r.seek(start, io.SEEK_SET)
    return hashlib.sha1(
        r.read(length)
    ).hexdigest().upper()


class TokenApi(api.ApiSpec):

    def __init__(self) -> None:
        super().__init__('https://uplb.115.com/3.0/gettoken.php')
    
    def parse_result(self, result: dict):
        if result.get('StatusCode' '') == '200':
            return {
                'access_key_id': result['AccessKeyId'],
                'access_key_secret': result['AccessKeySecret'],
                'security_token': result['SecurityToken'],
                'expiration': datetime.datetime.strptime(
                    result['Expiration'], '%Y-%m-%dT%H:%M:%SZ'
                )
            }
        return None
